@@ -1,6 +1,6 @@
+from __future__ import annotations
 from dataclasses import dataclass
 from xdsl.ir import *
-from xdsl.util import OpOrBlockArg, get_ssa_value
 from xdsl.dialects.builtin import *
 from xdsl.irdl import *
 
@@ -21,72 +21,9 @@ class MemRef:
         self.ctx.register_op(GetGlobal)
         self.ctx.register_op(Global)
 
-    def load(self,
-             value: OpOrBlockArg,
-             indices: List[OpOrBlockArg] = []) -> Operation:
-        # TODO should we really check these things here?
-        if not isinstance(get_ssa_value(value).typ, MemRefType):
-            raise Exception("memref.load expected a MemRefType operand")
-        return Load.create([get_ssa_value(value)] +
-                           [get_ssa_value(i) for i in indices],
-                           [get_ssa_value(value).typ.element_type])
-
-    def store(self,
-              value: OpOrBlockArg,
-              place: OpOrBlockArg,
-              indices: List[OpOrBlockArg] = []) -> Operation:
-        return Store.create(
-            [get_ssa_value(value), get_ssa_value(place)] +
-            [get_ssa_value(i) for i in indices], [])
-
-    def alloc(self,
-              alignment: int,
-              return_type: Attribute,
-              shape: List[int] = [1]) -> Operation:
-        return Alloc.create(
-            [], [MemRefType.get(return_type, shape)],
-            attributes={
-                "alignment": IntegerAttr.get(alignment, IntegerType.get(64)),
-                "operand_segment_sizes": VectorAttr.get([0, 0])
-            })
-
-    def alloca(self,
-               alignment: int,
-               return_type: Attribute,
-               shape: List[int] = [1]) -> Operation:
-        return Alloca.create(
-            [], [MemRefType.get(return_type, shape)],
-            attributes={
-                "alignment": IntegerAttr.get(alignment, IntegerType.get(64)),
-                "operand_segment_sizes": VectorAttr.get([0, 0])
-            })
-
-    def dealloc(self, memref: OpOrBlockArg) -> Operation:
-        return Dealloc.create([get_ssa_value(memref)], [])
-
-    def get_global(self, name: str, return_type: Attribute) -> Operation:
-        return GetGlobal.create([], [return_type],
-                                {"name": FlatSymbolRefAttr.get(name)})
-
-    def global_(self,
-                sym_name: str,
-                typ: Attribute,
-                initial_value: Optional[Attribute],
-                sym_visibility: str = "private",
-                constant: bool = False) -> Operation:
-        if not initial_value:
-            raise Exception("optional arguments are not yet supported")
-        return Global.create(
-            [], [], {
-                "sym_name": SymbolNameAttr.get(sym_name),
-                "type": typ,
-                "initial_value": initial_value,
-                "sym_visibility": StringAttr.get(sym_visibility)
-            })
-
 
 @irdl_attr_definition
-class MemRefType:
+class MemRefType(ParametrizedAttribute):
     name = "memref"
 
     shape = ParameterDef(ArrayOfConstraint(IntegerAttr))
@@ -99,16 +36,29 @@ class MemRefType:
         return [i.parameters[0].data for i in self.shape.data]
 
     @staticmethod
-    def get(referenced_type: Attribute, shape: List[int] = [1]) -> Attribute:
+    @builder
+    def from_type_and_list(
+            referenced_type: Attribute,
+            shape: List[Union[int, IntegerAttr]] = None) -> MemRefType:
+        if shape is None:
+            shape = [1]
         return MemRefType([
-            ArrayAttr.get(
-                [IntegerAttr.get(d, IntegerType.get(64)) for d in shape]),
+            ArrayAttr.from_list([IntegerAttr.build(d) for d in shape]),
             referenced_type
         ])
 
+    @staticmethod
+    @builder
+    def from_params(
+        referenced_type: Attribute,
+        shape: ArrayAttr = ArrayAttr.from_list(
+            [IntegerAttr.from_int_and_width(1, 64)])
+    ) -> MemRefType:
+        return MemRefType([shape, referenced_type])
+
 
 @irdl_op_definition
-class Load:
+class Load(Operation):
     name = "memref.load"
     memref = OperandDef(MemRefType)
     indices = VarOperandDef(IndexType)
@@ -125,9 +75,15 @@ class Load:
         if self.memref.typ.get_num_dims() != len(self.indices):
             raise Exception("expected an index for each dimension")
 
+    @staticmethod
+    def get(ref: Union[SSAValue, Operation],
+            indices: List[Union[SSAValue, Operation]]) -> Load:
+        return Load.build(operands=[ref, indices],
+                          result_types=[SSAValue.get(ref).typ.element_type])
+
 
 @irdl_op_definition
-class Store:
+class Store(Operation):
     name = "memref.store"
     value = OperandDef(AnyAttr())
     memref = OperandDef(MemRefType)
@@ -141,9 +97,14 @@ class Store:
         if self.memref.typ.get_num_dims() != len(self.indices):
             raise Exception("expected an index for each dimension")
 
+    @staticmethod
+    def get(value: Union[Operation, SSAValue], ref: Union[Operation, SSAValue],
+            indices: List[Union[Operation, SSAValue]]) -> Store:
+        return Store.build(operands=[value, ref, indices])
+
 
 @irdl_op_definition
-class Alloc:
+class Alloc(Operation):
     name = "memref.alloc"
 
     dynamic_sizes = VarOperandDef(IndexType)
@@ -156,9 +117,22 @@ class Alloc:
 
     irdl_options = [AttrSizedOperandSegments()]
 
+    @staticmethod
+    def get(return_type: Attribute,
+            alignment: int,
+            shape: List[Union[int, IntegerAttr]] = None) -> Alloc:
+        if shape is None:
+            shape = [1]
+        return Alloc.build(
+            operands=[[], []],
+            result_types=[MemRefType.from_type_and_list(return_type, shape)],
+            attributes={
+                "alignment": IntegerAttr.from_int_and_width(alignment, 64)
+            })
+
 
 @irdl_op_definition
-class Alloca:
+class Alloca(Operation):
     name = "memref.alloca"
 
     dynamic_sizes = VarOperandDef(IndexType)
@@ -171,14 +145,31 @@ class Alloca:
 
     irdl_options = [AttrSizedOperandSegments()]
 
+    @staticmethod
+    def get(return_type: Attribute,
+            alignment: int,
+            shape: List[Union[int, IntegerAttr]] = None) -> Alloca:
+        if shape is None:
+            shape = [1]
+        return Alloca.build(
+            operands=[[], []],
+            result_types=[MemRefType.from_type_and_list(return_type, shape)],
+            attributes={
+                "alignment": IntegerAttr.from_int_and_width(alignment, 64)
+            })
+
 
 @irdl_op_definition
-class Dealloc:
+class Dealloc(Operation):
     name = "memref.dealloc"
     memref = OperandDef(MemRefType)
 
+    @staticmethod
+    def get(operand: Union[Operation, SSAValue]) -> Dealloc:
+        return Dealloc.build(operands=[operand])
 
-#@irdl_op_definition
+
+@irdl_op_definition
 class GetGlobal(Operation):
     name = "memref.get_global"
     #name = AttributeDef(FlatSymbolRefAttr)
@@ -193,11 +184,17 @@ class GetGlobal(Operation):
             raise Exception(
                 "expected 'name' attribute to be a FlatSymbolRefAttr")
 
+    @staticmethod
+    def get(name, return_type: Attribute) -> GetGlobal:
+        return GetGlobal.build(
+            result_types=[return_type],
+            attributes={"name": FlatSymbolRefAttr.build(name)})
+
     #TODO how to verify the types, as the global might be defined in another compilation unit
 
 
 @irdl_op_definition
-class Global:
+class Global(Operation):
     name = "memref.global"
     sym_name = AttributeDef(SymbolNameAttr)
     sym_visibility = AttributeDef(StringAttr)
@@ -207,4 +204,17 @@ class Global:
     initial_value = AttributeDef(AnyAttr())
 
     # TODO how do we represent these in MLIR-Lite
-    #constant = AttributeDef(UnitAttr)
+    # constant = AttributeDef(UnitAttr)
+
+    @staticmethod
+    def get(sym_name,
+            typ: Attribute,
+            initial_value: Optional[Attribute],
+            sym_visibility="private") -> Global:
+        return Global.build(
+            attributes={
+                "sym_name": sym_name,
+                "type": typ,
+                "initial_value": initial_value,
+                "sym_visibility": sym_visibility
+            })
