@@ -1,3 +1,4 @@
+from __future__ import annotations
 from xdsl.ir import *
 from xdsl.irdl import *
 from xdsl.util import *
@@ -29,83 +30,36 @@ class Std:
 
         self.ctx.register_op(Cmpi)
 
-        self.f32 = Float32Type.get()
-        self.i64 = IntegerType.get(64)
-        self.i32 = IntegerType.get(32)
-        self.i1 = IntegerType.get(1)
-
-    # TODO make this generic in the type
-    def constant(self, val: int, typ: Attribute) -> Operation:
-        return Constant.create([], [typ],
-                               attributes={"value": IntegerAttr.get(val, typ)})
-
-    def constant_from_attr(self, attr: Attribute, typ: Attribute) -> Operation:
-        return Constant.create([], [typ], attributes={"value": attr})
-
-    def mulf(self, x: OpOrBlockArg, y: OpOrBlockArg) -> Operation:
-        return Mulf.create([get_ssa_value(x), get_ssa_value(y)], [self.f32])
-
-    def addf(self, x: OpOrBlockArg, y: OpOrBlockArg) -> Operation:
-        return Addf.create([get_ssa_value(x), get_ssa_value(y)], [self.f32])
-
-    def call(self, callee: str, ops: List[OpOrBlockArg],
-             return_types: List[Attribute]) -> Operation:
-        return Call.create(
-            [get_ssa_value(op) for op in ops],
-            return_types,
-            attributes={"callee": FlatSymbolRefAttr.get(callee)})
-
-    def return_(self, *ops: OpOrBlockArg) -> Operation:
-        return Return.create([get_ssa_value(op) for op in ops], [])
-
-    # TODO these operations should support all kinds of integer types
-    def muli(self, x: OpOrBlockArg, y: OpOrBlockArg) -> Operation:
-        return Muli.create([get_ssa_value(x), get_ssa_value(y)], [self.i32])
-
-    def addi(self, x: OpOrBlockArg, y: OpOrBlockArg) -> Operation:
-        return Addi.create([get_ssa_value(x), get_ssa_value(y)], [self.i32])
-
-    def subi(self, x: OpOrBlockArg, y: OpOrBlockArg) -> Operation:
-        return Subi.create([get_ssa_value(x), get_ssa_value(y)], [self.i32])
-
-    def floordivi_signed(self, x: OpOrBlockArg, y: OpOrBlockArg) -> Operation:
-        return FloordiviSigned.create(
-            [get_ssa_value(x), get_ssa_value(y)], [self.i32])
-
-    def remi_signed(self, x: OpOrBlockArg, y: OpOrBlockArg) -> Operation:
-        return RemiSigned.create(
-            [get_ssa_value(x), get_ssa_value(y)], [self.i32])
-
-    # TODO these operations should support all kinds of integer types
-    def and_(self, x: OpOrBlockArg, y: OpOrBlockArg) -> Operation:
-        return And.create([get_ssa_value(x), get_ssa_value(y)], [self.i1])
-
-    def or_(self, x: OpOrBlockArg, y: OpOrBlockArg) -> Operation:
-        return Or.create([get_ssa_value(x), get_ssa_value(y)], [self.i1])
-
-    def xor_(self, x: OpOrBlockArg, y: OpOrBlockArg) -> Operation:
-        return Xor.create([get_ssa_value(x), get_ssa_value(y)], [self.i1])
-
-    def cmpi(self, x: OpOrBlockArg, y: OpOrBlockArg, arg: int) -> Operation:
-        return Cmpi.create(
-            [get_ssa_value(x), get_ssa_value(y)], [self.i1],
-            attributes={"predicate": IntegerAttr.get(arg, self.i64)})
+        self.f32 = Float32Type()
+        self.i64 = IntegerType.from_width(64)
+        self.i32 = IntegerType.from_width(32)
+        self.i1 = IntegerType.from_width(1)
 
 
 @irdl_op_definition
-class Constant:
+class Constant(Operation):
     name: str = "std.constant"
     output = ResultDef(AnyAttr())
     value = AttributeDef(AnyAttr())
 
     # TODO verify that the output and value type are equal
-    def verify_(self) -> None:
-        # TODO how to force the attr to have a type? and how to query it?
-        pass
+
+    @staticmethod
+    def from_attr(attr: Attribute, typ: Attribute) -> Constant:
+        return Constant.create(result_types=[typ], attributes={"value": attr})
+
+    @staticmethod
+    def from_int_constant(val: Union[int, Attribute],
+                          typ: Union[int, Attribute]) -> Constant:
+        if isinstance(typ, int):
+            typ = IntegerType.from_width(typ)
+        return Constant.create(
+            result_types=[typ],
+            attributes={"value": IntegerAttr.from_params(val, typ)})
 
 
 @irdl_op_definition
-class Addi:
+class Addi(Operation):
     name: str = "std.addi"
     input1 = OperandDef(IntegerType)
     input2 = OperandDef(IntegerType)
@@ -116,9 +70,15 @@ class Addi:
         if self.input1.typ != self.input2.typ or self.input2.typ != self.output.typ:
             raise Exception("expect all input and output types to be equal")
 
+    @staticmethod
+    def get(operand1: Union[Operation, SSAValue],
+            operand2: Union[Operation, SSAValue]) -> Addi:
+        return Addi.build(operands=[operand1, operand2],
+                          result_types=[IntegerType.from_width(32)])
+
 
 @irdl_op_definition
-class Muli:
+class Muli(Operation):
     name: str = "std.muli"
     input1 = OperandDef(IntegerType)
     input2 = OperandDef(IntegerType)
@@ -129,9 +89,15 @@ class Muli:
         if self.input1.typ != self.input2.typ or self.input2.typ != self.output.typ:
             raise Exception("expect all input and output types to be equal")
 
+    @staticmethod
+    def get(operand1: Union[Operation, SSAValue],
+            operand2: Union[Operation, SSAValue]) -> Muli:
+        return Muli.build(operands=[operand1, operand2],
+                          result_types=[IntegerType.from_width(32)])
+
 
 @irdl_op_definition
-class Subi:
+class Subi(Operation):
     name: str = "std.subi"
     input1 = OperandDef(IntegerType)
     input2 = OperandDef(IntegerType)
@@ -142,9 +108,15 @@ class Subi:
         if self.input1.typ != self.input2.typ or self.input2.typ != self.output.typ:
             raise Exception("expect all input and output types to be equal")
 
+    @staticmethod
+    def get(operand1: Union[Operation, SSAValue],
+            operand2: Union[Operation, SSAValue]) -> Subi:
+        return Subi.build(operands=[operand1, operand2],
+                          result_types=[IntegerType.from_width(32)])
+
 
 @irdl_op_definition
-class FloordiviSigned:
+class FloordiviSigned(Operation):
     name: str = "std.floordivi_signed"
     input1 = OperandDef(IntegerType)
     input2 = OperandDef(IntegerType)
@@ -155,9 +127,15 @@ class FloordiviSigned:
         if self.input1.typ != self.input2.typ or self.input2.typ != self.output.typ:
             raise Exception("expect all input and output types to be equal")
 
+    @staticmethod
+    def get(operand1: Union[Operation, SSAValue],
+            operand2: Union[Operation, SSAValue]) -> FloordiviSigned:
+        return FloordiviSigned.build(operands=[operand1, operand2],
+                                     result_types=[IntegerType.from_width(32)])
+
 
 @irdl_op_definition
-class RemiSigned:
+class RemiSigned(Operation):
     name: str = "std.remi_signed"
     input1 = OperandDef(IntegerType)
     input2 = OperandDef(IntegerType)
@@ -168,9 +146,15 @@ class RemiSigned:
         if self.input1.typ != self.input2.typ or self.input2.typ != self.output.typ:
             raise Exception("expect all input and output types to be equal")
 
+    @staticmethod
+    def get(operand1: Union[Operation, SSAValue],
+            operand2: Union[Operation, SSAValue]) -> RemiSigned:
+        return RemiSigned.build(operands=[operand1, operand2],
+                                result_types=[IntegerType.from_width(32)])
+
 
 @irdl_op_definition
-class Call:
+class Call(Operation):
     name: str = "std.call"
     arguments = VarOperandDef(AnyAttr())
     callee = AttributeDef(FlatSymbolRefAttr)
@@ -179,15 +163,27 @@ class Call:
     res = VarResultDef(AnyAttr())
     # TODO how do we verify that the types are correct?
 
+    @staticmethod
+    def get(callee: Union[str, FlatSymbolRefAttr],
+            operands: List[Union[SSAValue, Operation]],
+            return_types: List[Attribute]) -> Call:
+        return Call.build(operands=operands,
+                          result_types=return_types,
+                          attributes={"callee": callee})
+
 
 @irdl_op_definition
-class Return:
+class Return(Operation):
     name: str = "std.return"
     arguments = VarOperandDef(AnyAttr())
 
+    @staticmethod
+    def get(*ops: Union[Operation, SSAValue]) -> Return:
+        return Return.build(operands=[[op for op in ops]])
+
 
 @irdl_op_definition
-class And:
+class And(Operation):
     name: str = "std.and"
     input1 = OperandDef(IntegerType)
     input2 = OperandDef(IntegerType)
@@ -198,9 +194,15 @@ class And:
         if self.input1.typ != self.input2.typ or self.input2.typ != self.output.typ:
             raise Exception("expect all input and output types to be equal")
 
+    @staticmethod
+    def get(operand1: Union[Operation, SSAValue],
+            operand2: Union[Operation, SSAValue]) -> And:
+        return And.build(operands=[operand1, operand2],
+                         result_types=[IntegerType.from_width(1)])
+
 
 @irdl_op_definition
-class Or:
+class Or(Operation):
     name: str = "std.or"
     input1 = OperandDef(IntegerType)
     input2 = OperandDef(IntegerType)
@@ -211,9 +213,15 @@ class Or:
         if self.input1.typ != self.input2.typ or self.input2.typ != self.output.typ:
             raise Exception("expect all input and output types to be equal")
 
+    @staticmethod
+    def get(operand1: Union[Operation, SSAValue],
+            operand2: Union[Operation, SSAValue]) -> Or:
+        return Or.build(operands=[operand1, operand2],
+                        result_types=[IntegerType.from_width(1)])
+
 
 @irdl_op_definition
-class Xor:
+class Xor(Operation):
     name: str = "std.xor"
     input1 = OperandDef(IntegerType)
     input2 = OperandDef(IntegerType)
@@ -224,18 +232,32 @@ class Xor:
         if self.input1.typ != self.input2.typ or self.input2.typ != self.output.typ:
             raise Exception("expect all input and output types to be equal")
 
+    @staticmethod
+    def get(operand1: Union[Operation, SSAValue],
+            operand2: Union[Operation, SSAValue]) -> Xor:
+        return Xor.build(operands=[operand1, operand2],
+                         result_types=[IntegerType.from_width(1)])
+
 
 @irdl_op_definition
-class Cmpi:
+class Cmpi(Operation):
     name: str = "std.cmpi"
     predicate = AttributeDef(IntegerAttr)
     input1 = OperandDef(IntegerType)
     input2 = OperandDef(IntegerType)
-    output = ResultDef(IntegerType.get(1))
+    output = ResultDef(IntegerType.from_width(1))
+
+    @staticmethod
+    def get(operand1: Union[Operation, SSAValue],
+            operand2: Union[Operation, SSAValue], arg: int) -> Cmpi:
+        return Cmpi.build(
+            operands=[operand1, operand2],
+            result_types=[IntegerType.from_width(1)],
+            attributes={"predicate": IntegerAttr.from_int_and_width(arg, 64)})
 
 
 @irdl_op_definition
-class Addf:
+class Addf(Operation):
     name: str = "std.addf"
     input1 = OperandDef(Float32Type)
     input2 = OperandDef(Float32Type)
@@ -246,9 +268,15 @@ class Addf:
         if self.input1.typ != self.input2.typ or self.input2.typ != self.output.typ:
             raise Exception("expect all input and output types to be equal")
 
+    @staticmethod
+    def get(operand1: Union[Operation, SSAValue],
+            operand2: Union[Operation, SSAValue]) -> Addf:
+        return Addf.build(operands=[operand1, operand2],
+                          result_types=[Float32Type()])
+
 
 @irdl_op_definition
-class Mulf:
+class Mulf(Operation):
     name: str = "std.mulf"
     input1 = OperandDef(Float32Type)
     input2 = OperandDef(Float32Type)
@@ -258,3 +286,9 @@ class Mulf:
     def verify_(self) -> None:
         if self.input1.typ != self.input2.typ or self.input2.typ != self.output.typ:
             raise Exception("expect all input and output types to be equal")
+
+    @staticmethod
+    def get(operand1: Union[Operation, SSAValue],
+            operand2: Union[Operation, SSAValue]) -> Mulf:
+        return Mulf.build(operands=[operand1, operand2],
+                          result_types=[Float32Type()])
