@@ -1,16 +1,20 @@
 from __future__ import annotations
+
+import inspect
 from dataclasses import dataclass
 from abc import ABC, abstractmethod
-
-from xdsl.ir import Operation, Attribute, ParametrizedAttribute, SSAValue
-from typing import List, Tuple, Optional, Union
+from typing import List, Tuple, Optional, Union, TypeVar
 from inspect import isclass
 import typing
 
+from xdsl.ir import Operation, Attribute, ParametrizedAttribute, SSAValue, Data, Region
+from xdsl import util
+
 
 @dataclass
 class AttrConstraint(ABC):
     """Constrain an attribute to a certain value."""
+
     @abstractmethod
     def verify(self, attr: Attribute) -> None:
         """
@@ -26,6 +30,7 @@ class EqAttrConstraint(AttrConstraint):
 
     attr: Attribute
     """The attribute we want to check equality with."""
+
     def verify(self, attr: Attribute) -> None:
         if attr != self.attr:
             raise Exception(f"Expected attribute {self.attr} but got {attr}")
@@ -37,6 +42,7 @@ class BaseAttr(AttrConstraint):
 
     attr: typing.Type[Attribute]
     """The expected attribute base type."""
+
     def verify(self, attr: Attribute) -> None:
         if not isinstance(attr, self.attr):
             raise Exception(
@@ -60,8 +66,10 @@ def attr_constr_coercion(
 @dataclass
 class AnyAttr(AttrConstraint):
     """Constraint that is verified by all attributes."""
+
     def verify(self, attr: Attribute) -> None:
-        pass
+        if not isinstance(attr, Attribute):
+            raise Exception(f"Expected attribute, but got {attr}")
 
 
 @dataclass(init=False)
@@ -70,6 +78,7 @@ class AnyOf(AttrConstraint):
 
     attr_constrs: List[AttrConstraint]
     """The list of constraints that are checked."""
+
     def __init__(self,
                  attr_constrs: List[Union[Attribute, typing.Type[Attribute],
                                           AttrConstraint]]):
@@ -101,6 +110,7 @@ class ParamAttrConstraint(AttrConstraint):
 
     param_constrs: List[AttrConstraint]
     """The attribute parameter constraints"""
+
     def __init__(self, base_attr: typing.Type[Attribute],
                  param_constrs: List[Union[Attribute, typing.Type[Attribute],
                                            AttrConstraint]]):
@@ -169,6 +179,7 @@ class OperandDef(OperandOrResultDef):
 
     constr: AttrConstraint
     """The operand constraint."""
+
     def __init__(self, typ: Union[Attribute, typing.Type[Attribute],
                                   AttrConstraint]):
         self.constr = attr_constr_coercion(typ)
@@ -190,6 +201,7 @@ class ResultDef(OperandOrResultDef):
 
     constr: AttrConstraint
     """The result constraint."""
+
     def __init__(self, typ: Union[Attribute, typing.Type[Attribute],
                                   AttrConstraint]):
         self.constr = attr_constr_coercion(typ)
@@ -226,6 +238,7 @@ class AttributeDef:
 
     constr: AttrConstraint
     """The attribute constraint."""
+
     def __init__(self, typ: Union[Attribute, typing.Type[Attribute],
                                   AttrConstraint]):
         self.constr = attr_constr_coercion(typ)
@@ -345,7 +358,6 @@ def irdl_op_verify(op: Operation, operands: List[Tuple[str, OperandDef]],
             for i in range(operand_sizes[current_var_operand]):
                 operand_def.constr.verify(op.operands[current_operand].typ)
                 current_operand += 1
-            current_operand += 1
         else:
             operand_def.constr.verify(op.operands[current_operand].typ)
             current_operand += 1
@@ -360,7 +372,6 @@ def irdl_op_verify(op: Operation, operands: List[Tuple[str, OperandDef]],
             for i in range(result_sizes[current_var_result]):
                 result_def.constr.verify(op.results[current_result].typ)
                 current_result += 1
-            current_result += 1
         else:
             result_def.constr.verify(op.results[current_result].typ)
             current_result += 1
@@ -401,26 +412,137 @@ def irdl_op_verify(op: Operation, operands: List[Tuple[str, OperandDef]],
         attr_def.constr.verify(op.attributes[attr_name])
 
 
-def irdl_op_definition(cls) -> typing.Type[Operation]:
+def irdl_build_attribute(irdl_def: AttrConstraint, result) -> Attribute:
+    if isinstance(irdl_def, BaseAttr):
+        if isinstance(result, Tuple):
+            return irdl_def.attr.build(*result)
+        return irdl_def.attr.build(result)
+    elif irdl_def == AnyAttr():
+        if isinstance(result, Attribute):
+            return result
+        raise Exception(f"builder expected an attribute, got {result}")
+    else:
+        raise NotImplementedError(
+            f"Builder does not support IRDL definition {irdl_def}")
+
+
+OpT = TypeVar('OpT', bound='Operation')
+
+
+def irdl_op_builder(cls: typing.Type[OpT], operands: List,
+                    operand_defs: List[Tuple[str, OperandDef]],
+                    res_types: List, res_defs: List[Tuple[str, ResultDef]],
+                    attributes: typing.Dict[str, typing.Any],
+                    attr_defs: typing.Dict[str, AttributeDef], successors,
+                    regions, options) -> OpT:
+    """Builder for an irdl operation."""
+
+    # We need irdl to define VectorAttr, but here we need VectorAttr.
+    # So we have a circular dependency that we solve by importing in this function.
+    from xdsl.dialects.builtin import VectorAttr, IntegerAttr
+
+    # Build operands by forwarding the values to SSAValue.get
+    if len(operand_defs) != len(operands):
+        raise ValueError(
+            f"Expected {len(operand_defs)} operands, got {len(operands)}")
+
+    built_operands = []
+    for ((_, operand_def), operand) in zip(operand_defs, operands):
+        if isinstance(operand_def, VarOperandDef):
+            if not isinstance(operand, list):
+                raise ValueError(
+                    f"Expected list for variadic operand builder, got {operand}"
+                )
+            built_operands.extend([SSAValue.get(arg) for arg in operand])
+        else:
+            built_operands.append(SSAValue.get(operand))
+
+    # Build results by forwarding the values to the attribute builders
+    if len(res_defs) != len(res_types):
+        raise ValueError(
+            f"Expected {len(res_defs)} results, got {len(res_types)}")
+
+    built_res_types = []
+    for ((_, res_def), res_type) in zip(res_defs, res_types):
+        if isinstance(res_def, VarResultDef):
+            if not isinstance(res_type, list):
+                raise ValueError(
+                    f"Expected list for variadic result builder, got {res_type}"
+                )
+            built_res_types.extend([
+                irdl_build_attribute(res_def.constr, res) for res in res_type
+            ])
+        else:
+            built_res_types.append(
+                irdl_build_attribute(res_def.constr, res_type))
+
+    # Build attributes by forwarding the values to the attribute builders
+    attr_defs = {name: def_ for (name, def_) in attr_defs}
+    built_attributes = dict()
+    for attr_name, attr in attributes.items():
+        if attr_name not in attr_defs:
+            if isinstance(attr, Attribute):
+                built_attributes[attr_name] = attr
+                continue
+            raise ValueError(
+                f"Unexpected attribute name {attr_name} for operation {cls.name}"
+            )
+        built_attributes[attr_name] = irdl_build_attribute(
+            attr_defs[attr_name].constr, attr)
+
+    # Take care of variadic operand and result segment sizes.
+    if AttrSizedOperandSegments() in options:
+        sizes = [
+            len(operand)
+            for operand, (_, operand_def) in zip(operands, operand_defs)
+            if isinstance(operand_def, VarOperandDef)
+        ]
+        built_attributes[AttrSizedOperandSegments.attribute_name] =\
+            VectorAttr.from_list([IntegerAttr.from_index_int_value(size) for size in sizes])
+
+    if AttrSizedResultSegments() in options:
+        sizes = [
+            len(result)
+            for result, (_, result_def) in zip(res_types, res_defs)
+            if isinstance(result_def, VarResultDef)
+        ]
+        built_attributes[AttrSizedResultSegments.attribute_name] =\
+            VectorAttr.from_list([IntegerAttr.from_index_int_value(size) for size in sizes])
+
+    # Build regions using `Region.get`.
+    regions = [Region.get(region) for region in regions]
+
+    return cls.create(operands=built_operands,
+                      result_types=built_res_types,
+                      attributes=built_attributes,
+                      successors=successors,
+                      regions=regions)
+
+
+OperationType = TypeVar("OperationType", bound=Operation)
+
+
+def irdl_op_definition(
+        cls: typing.Type[OperationType]) -> typing.Type[OperationType]:
     """Decorator used on classes to define a new operation definition."""
-    operands = [(field_name, field)
-                for field_name, field in cls.__dict__.items()
-                if isinstance(field, OperandDef)]
-    results = [(field_name, field)
-               for field_name, field in cls.__dict__.items()
-               if isinstance(field, ResultDef)]
-    regions = [(field_name, field)
-               for field_name, field in cls.__dict__.items()
-               if isinstance(field, RegionDef)]
-    attributes = [(field_name, field)
-                  for field_name, field in cls.__dict__.items()
-                  if isinstance(field, AttributeDef)]
+    operand_defs = [(field_name, field)
+                    for field_name, field in cls.__dict__.items()
+                    if isinstance(field, OperandDef)]
+    result_defs = [(field_name, field)
+                   for field_name, field in cls.__dict__.items()
+                   if isinstance(field, ResultDef)]
+    region_defs = [(field_name, field)
+                   for field_name, field in cls.__dict__.items()
+                   if isinstance(field, RegionDef)]
+    attr_defs = [(field_name, field)
+                 for field_name, field in cls.__dict__.items()
+                 if isinstance(field, AttributeDef)]
     options = cls.__dict__.get("irdl_options", [])
     new_attrs = dict()
 
     # Add operand access fields
     previous_variadics = 0
-    for operand_idx, (operand_name, operand_def) in enumerate(operands):
+    for operand_idx, (operand_name, operand_def) in enumerate(operand_defs):
         new_attrs[operand_name] = property(
             lambda self, idx=operand_idx, previous_vars=previous_variadics:
             get_operand_or_result(self, idx, previous_vars, True))
@@ -433,7 +555,7 @@ def irdl_op_definition(cls) -> typing.Type[Operation]:
 
     # Add result access fields
     previous_variadics = 0
-    for result_idx, (result_name, result_def) in enumerate(results):
+    for result_idx, (result_name, result_def) in enumerate(result_defs):
         new_attrs[result_name] = property(
             lambda self, idx=result_idx, previous_vars=previous_variadics:
             get_operand_or_result(self, idx, previous_vars, False))
@@ -443,22 +565,22 @@ def irdl_op_definition(cls) -> typing.Type[Operation]:
         raise Exception("Operation defines more than two variadic results, "
                         "but do not define the AttrSizedResultSegments option")
 
-    for region_idx, (region_name, _) in enumerate(regions):
+    for region_idx, (region_name, _) in enumerate(region_defs):
         new_attrs[region_name] = property(
             (lambda idx: lambda self: self.regions[idx])(region_idx))
 
-    for attribute_name, _ in attributes:
+    for attribute_name, _ in attr_defs:
         new_attrs[attribute_name] = property(
             (lambda name: lambda self: self.attributes[name])(attribute_name))
 
-    new_attrs["irdl_operand_defs"] = operands
-    new_attrs["irdl_result_defs"] = results
-    new_attrs["irdl_region_defs"] = regions
-    new_attrs["irdl_attribute_defs"] = attributes
+    new_attrs["irdl_operand_defs"] = operand_defs
+    new_attrs["irdl_result_defs"] = result_defs
+    new_attrs["irdl_region_defs"] = region_defs
+    new_attrs["irdl_attribute_defs"] = attr_defs
     new_attrs["irdl_options"] = options
 
-    new_attrs["verify_"] = lambda op: irdl_op_verify(op, operands, results,
-                                                     regions, attributes)
+    new_attrs["verify_"] = lambda op: irdl_op_verify(
+        op, operand_defs, result_defs, region_defs, attr_defs)
     if "verify_" in cls.__dict__:
         custom_verifier = cls.__dict__["verify_"]
 
@@ -470,6 +592,18 @@ def irdl_op_definition(cls) -> typing.Type[Operation]:
             lambda verifier: lambda op: new_verifier(verifier, op))(
                 new_attrs["verify_"])
 
+    def builder(cls,
+                operands=[],
+                result_types=[],
+                attributes=dict(),
+                successors=[],
+                regions=[]):
+        return irdl_op_builder(cls, operands, operand_defs, result_types,
+                               result_defs, attributes, attr_defs, successors,
+                               regions, options)
+
+    new_attrs["build"] = classmethod(builder)
+
     return type(cls.__name__, (Operation, ), {**cls.__dict__, **new_attrs})
 
 
@@ -495,7 +629,77 @@ def irdl_attr_verify(attr: ParametrizedAttribute,
         param_def.constr.verify(attr.parameters[idx])
 
 
-def irdl_attr_definition(cls) -> typing.Type[Attribute]:
+C = TypeVar('C', bound='Callable')
+
+
+def builder(f: C) -> C:
+    """
+    Annotate a function and mark it as an IRDL builder.
+    This should only be used as decorator in classes decorated by irdl_attr_builder.
+    """
+    f.__irdl_is_builder = True
+    return f
+
+
+def irdl_get_builders(cls) -> List[typing.Callable]:
+    builders = []
+    for field_name in cls.__dict__:
+        field_ = cls.__dict__[field_name]
+        # Builders are staticmethods, so we need to get back the original function with __func__
+        if hasattr(field_, "__func__") and hasattr(field_.__func__,
+                                                   "__irdl_is_builder"):
+            builders.append(field_.__func__)
+    return builders
+
+
+def irdl_attr_try_builder(builder, *args):
+    params_dict = typing.get_type_hints(builder)
+    builder_params = inspect.signature(builder).parameters
+    params = [params_dict[param.name] for param in builder_params.values()]
+    defaults = [param.default for param in builder_params.values()]
+    num_non_defaults = defaults.count(inspect.Signature.empty)
+    if num_non_defaults > len(args):
+        return None
+    for arg, param in zip(args, params[:num_non_defaults]):
+        if not util.is_satisfying_hint(arg, param):
+            return None
+    return builder(*args, *defaults[len(args):])
+
+
+def irdl_attr_builder(cls, builders, *args):
+    """Try to apply all builders to construct an attribute instance."""
+    if len(args) == 1 and isinstance(args[0], cls):
+        return args[0]
+    for builder in builders:
+        res = irdl_attr_try_builder(builder, *args)
+        if res is not None:
+            return res
+    raise TypeError(
+        f"No available {cls.__name__} builders for arguments {args}")
+
+
+T = TypeVar('T')
+
+
+def irdl_data_definition(cls: typing.Type[T]) -> typing.Type[T]:
+    builders = irdl_get_builders(cls)
+    if "build" in cls.__dict__:
+        raise Exception(
+            f'"build" method for {cls.__name__} is reserved for IRDL, and should not be defined.'
+        )
+    new_attrs = dict()
+    new_attrs["build"] = lambda *args: irdl_attr_builder(cls, builders, *args)
+    return dataclass(frozen=True)(type(cls.__name__, (cls, ), {
+        **cls.__dict__,
+        **new_attrs
+    }))
+
+
+AttributeType = TypeVar("AttributeType", bound=ParametrizedAttribute)
+
+
+def irdl_param_attr_definition(
+        cls: typing.Type[AttributeType]) -> typing.Type[AttributeType]:
     """Decorator used on classes to define a new attribute definition."""
 
     parameters = []
@@ -521,7 +725,24 @@ def irdl_attr_definition(cls) -> typing.Type[Attribute]:
             lambda verifier: lambda op: new_verifier(verifier, op))(
                 new_attrs["verify"])
 
-    return type(cls.__name__, (ParametrizedAttribute, ), {
+    builders = irdl_get_builders(cls)
+    if "build" in cls.__dict__:
+        raise Exception(
+            f'"build" method for {cls.__name__} is reserved for IRDL, and should not be defined.'
+        )
+    new_attrs["build"] = lambda *args: irdl_attr_builder(cls, builders, *args)
+
+    return dataclass(frozen=True)(type(cls.__name__, (cls, ), {
         **cls.__dict__,
         **new_attrs
-    })
+    }))
+
+
+def irdl_attr_definition(cls: typing.Type[T]) -> typing.Type[T]:
+    if issubclass(cls, ParametrizedAttribute):
+        return irdl_param_attr_definition(cls)
+    if issubclass(cls, Data):
+        return irdl_data_definition(cls)
+    raise Exception(
+        f"Class {cls.__name__} should either be a subclass of 'Data' or 'ParametrizedAttribute'"
+    )
